#include "assemble.hh"
#include "aslexer_wrap.hh"

#include <cctype>
#include <cstdio>
#include <cstring>
#include <iostream>

void assembler::add_value(const char *value)
{
    if (*value == '.')
    {
        need_sym sym = { memory.size(), value };
        needed_symbols.push_back(sym);
        memory.push_back(0);
    }
    else if (*value == '-')
    {
        memory.push_back((0U-atoword(value+1)) & UWORD_MAX);
    }
    else
    {
        memory.push_back(atoword(value));
    }
}

bool assembler::assemble(FILE *fasm)
{
    aslexer scanner;
    const char *instn;
    
    scanner.set_in(fasm);
    
    while (true)
    {
        switch (scanner.lex())
        {
            case TK_ERROR:
                return false;
                
            case TK_END:
                return true;
                
            case TK_INSTN:
                instn = scanner.get_text();
                if (strcmp(instn, "NOP") == 0)
                {
                    /* NOP */
                    memory.push_back(MI_NOP);
                }
                else if (strcmp(instn, "HALT") == 0)
                {
                    /* HALT */
                    memory.push_back(MI_HALT);
                }
                else if (strcmp(instn, "JZ") == 0)
                {
                    /* JZ addr */
                    memory.push_back(MI_JZ);
                    if (scanner.lex() != TK_VALUE)
                        goto syntax_error;
                    add_value(scanner.get_text());
                }
                else if (strcmp(instn, "JNZ") == 0)
                {
                    /* JNZ addr */
                    memory.push_back(MI_JNZ);
                    if (scanner.lex() != TK_VALUE)
                        goto syntax_error;
                    add_value(scanner.get_text());
                }
                else if (strcmp(instn, "JMP") == 0)
                {
                    /* JMP addr */
                    /* JMP A */
                    /* JMP D */
                    switch (scanner.lex())
                    {
                        case TK_VALUE:
                            memory.push_back(MI_JMP_I);
                            add_value(scanner.get_text());
                            break;
                        case TK_REG_A:
                            memory.push_back(MI_JMP_A);
                            break;
                        case TK_REG_D:
                            memory.push_back(MI_JMP_D);
                            break;
                        default:
                            goto syntax_error;
                    }
                }
                else if (strcmp(instn, "CALL") == 0)
                {
                    /* CALL addr */
                    /* CALL A */
                    /* CALL D */
                    switch (scanner.lex())
                    {
                        case TK_VALUE:
                            memory.push_back(MI_CALL_I);
                            add_value(scanner.get_text());
                            break;
                        case TK_REG_A:
                            memory.push_back(MI_CALL_A);
                            break;
                        case TK_REG_D:
                            memory.push_back(MI_CALL_D);
                            break;
                        default:
                            goto syntax_error;
                    }
                }
                else if (strcmp(instn, "RET") == 0)
                {
                    /* RET */
                    memory.push_back(MI_RET);
                }
                else if (strcmp(instn, "LD") == 0)
                {
                    /* LD A val */
                    /* LD A D */
                    /* LD A (D) */
                    /* LD A Z */
                    /* LD D val */
                    /* LD D A */
                    /* LD D S+off */
                    /* LD D S-off */
                    /* LD D (D) */
                    /* LD (D) A */
                    switch (scanner.lex())
                    {
                        case TK_REG_A:
                            switch (scanner.lex())
                            {
                                case TK_VALUE:
                                    memory.push_back(MI_LD_A_I);
                                    add_value(scanner.get_text());
                                    break;
                                case TK_REG_D:
                                    memory.push_back(MI_LD_A_D);
                                    break;
                                case TK_REG_D_DEREF:
                                    memory.push_back(MI_LD_A_DD);
                                    break;
                                case TK_FLAG_Z:
                                    memory.push_back(MI_LD_A_Z);
                                    break;
                                default:
                                    goto syntax_error;
                            }
                            break;
                        case TK_REG_D:
                            switch (scanner.lex())
                            {
                                case TK_VALUE:
                                    memory.push_back(MI_LD_D_I);
                                    add_value(scanner.get_text());
                                    break;
                                case TK_REG_A:
                                    memory.push_back(MI_LD_D_A);
                                    break;
                                case TK_REG_S_PLUS:
                                    if (scanner.lex() != TK_VALUE || *scanner.get_text() == '.')
                                        goto syntax_error;
                                    memory.push_back(MI_LD_D_SI);
                                    memory.push_back(atoword(scanner.get_text()));
                                    break;
                                case TK_REG_S_MINUS:
                                    if (scanner.lex() != TK_VALUE || *scanner.get_text() == '.')
                                        goto syntax_error;
                                    memory.push_back(MI_LD_D_SI);
                                    memory.push_back((0U-atoword(scanner.get_text())) & UWORD_MAX);
                                    break;
                                case TK_REG_D_DEREF:
                                    memory.push_back(MI_LD_D_DD);
                                    break;
                                default:
                                    goto syntax_error;
                            }
                            break;
                        case TK_REG_D_DEREF:
                            switch (scanner.lex())
                            {
                                case TK_REG_A:
                                    memory.push_back(MI_LD_DD_A);
                                    break;
                                default:
                                    goto syntax_error;
                            }
                            break;
                        default:
                            goto syntax_error;
                    }
                }
                else if (strcmp(instn, "ADD") == 0)
                {
                    /* ADD A D */
                    /* ADD S off */
                    switch (scanner.lex())
                    {
                        case TK_REG_A:
                            if (scanner.lex() != TK_REG_D)
                                goto syntax_error;
                            memory.push_back(MI_ADD);
                            break;
                        case TK_REG_S:
                            if (scanner.lex() != TK_VALUE)
                                goto syntax_error;
                            memory.push_back(MI_ADD_S);
                            add_value(scanner.get_text());
                            break;
                        default:
                            goto syntax_error;
                    }
                }
                else if (strcmp(instn, "SUB") == 0)
                {
                    /* SUB A D */
                    if (scanner.lex() != TK_REG_A)
                        goto syntax_error;
                    if (scanner.lex() != TK_REG_D)
                        goto syntax_error;
                    memory.push_back(MI_SUB);
                }
                else if (strcmp(instn, "MUL") == 0)
                {
                    /* MUL A D */
                    if (scanner.lex() != TK_REG_A)
                        goto syntax_error;
                    if (scanner.lex() != TK_REG_D)
                        goto syntax_error;
                    memory.push_back(MI_MUL);
                }
                else if (strcmp(instn, "DIV") == 0)
                {
                    /* DIV A D */
                    if (scanner.lex() != TK_REG_A)
                        goto syntax_error;
                    if (scanner.lex() != TK_REG_D)
                        goto syntax_error;
                    memory.push_back(MI_DIV);
                }
                else if (strcmp(instn, "MOD") == 0)
                {
                    /* MOD A D */
                    if (scanner.lex() != TK_REG_A)
                        goto syntax_error;
                    if (scanner.lex() != TK_REG_D)
                        goto syntax_error;
                    memory.push_back(MI_MOD);
                }
                else if (strcmp(instn, "NEG") == 0)
                {
                    /* NEG A */
                    if (scanner.lex() != TK_REG_A)
                        goto syntax_error;
                    memory.push_back(MI_NEG);
                }
                else if (strcmp(instn, "INC") == 0)
                {
                    /* INC A */
                    if (scanner.lex() != TK_REG_A)
                        goto syntax_error;
                    memory.push_back(MI_INC);
                }
                else if (strcmp(instn, "DEC") == 0)
                {
                    /* DEC A */
                    if (scanner.lex() != TK_REG_A)
                        goto syntax_error;
                    memory.push_back(MI_DEC);
                }
                else if (strcmp(instn, "LT") == 0)
                {
                    /* LT A D */
                    if (scanner.lex() != TK_REG_A)
                        goto syntax_error;
                    if (scanner.lex() != TK_REG_D)
                        goto syntax_error;
                    memory.push_back(MI_LT);
                }
                else if (strcmp(instn, "LEQ") == 0)
                {
                    /* LEQ A D */
                    if (scanner.lex() != TK_REG_A)
                        goto syntax_error;
                    if (scanner.lex() != TK_REG_D)
                        goto syntax_error;
                    memory.push_back(MI_LEQ);
                }
                else if (strcmp(instn, "GT") == 0)
                {
                    /* GT A D */
                    if (scanner.lex() != TK_REG_A)
                        goto syntax_error;
                    if (scanner.lex() != TK_REG_D)
                        goto syntax_error;
                    memory.push_back(MI_GT);
                }
                else if (strcmp(instn, "GEQ") == 0)
                {
                    /* GEQ A D */
                    if (scanner.lex() != TK_REG_A)
                        goto syntax_error;
                    if (scanner.lex() != TK_REG_D)
                        goto syntax_error;
                    memory.push_back(MI_GEQ);
                }
                else if (strcmp(instn, "EQ") == 0)
                {
                    /* EQ A D */
                    if (scanner.lex() != TK_REG_A)
                        goto syntax_error;
                    if (scanner.lex() != TK_REG_D)
                        goto syntax_error;
                    memory.push_back(MI_EQ);
                }
                else if (strcmp(instn, "NEQ") == 0)
                {
                    /* NEQ A D */
                    if (scanner.lex() != TK_REG_A)
                        goto syntax_error;
                    if (scanner.lex() != TK_REG_D)
                        goto syntax_error;
                    memory.push_back(MI_NEQ);
                }
                else if (strcmp(instn, "AND") == 0)
                {
                    /* AND A D */
                    if (scanner.lex() != TK_REG_A)
                        goto syntax_error;
                    if (scanner.lex() != TK_REG_D)
                        goto syntax_error;
                    memory.push_back(MI_AND);
                }
                else if (strcmp(instn, "OR") == 0)
                {
                    /* OR A D */
                    if (scanner.lex() != TK_REG_A)
                        goto syntax_error;
                    if (scanner.lex() != TK_REG_D)
                        goto syntax_error;
                    memory.push_back(MI_OR);
                }
                else if (strcmp(instn, "XOR") == 0)
                {
                    /* XOR A D */
                    if (scanner.lex() != TK_REG_A)
                        goto syntax_error;
                    if (scanner.lex() != TK_REG_D)
                        goto syntax_error;
                    memory.push_back(MI_XOR);
                }
                else if (strcmp(instn, "LS") == 0)
                {
                    /* LS A D */
                    if (scanner.lex() != TK_REG_A)
                        goto syntax_error;
                    if (scanner.lex() != TK_REG_D)
                        goto syntax_error;
                    memory.push_back(MI_LS);
                }
                else if (strcmp(instn, "RS") == 0)
                {
                    /* RS A D */
                    if (scanner.lex() != TK_REG_A)
                        goto syntax_error;
                    if (scanner.lex() != TK_REG_D)
                        goto syntax_error;
                    memory.push_back(MI_RS);
                }
                else if (strcmp(instn, "CPL") == 0)
                {
                    /* CPL A */
                    if (scanner.lex() != TK_REG_A)
                        goto syntax_error;
                    memory.push_back(MI_CPL);
                }
                else if (strcmp(instn, "IN") == 0)
                {
                    /* IN A */
                    if (scanner.lex() != TK_REG_A)
                        goto syntax_error;
                    memory.push_back(MI_IN);
                }
                else if (strcmp(instn, "OUT") == 0)
                {
                    /* OUT A */
                    if (scanner.lex() != TK_REG_A)
                        goto syntax_error;
                    memory.push_back(MI_OUT);
                }
                else if (strcmp(instn, "LBL") == 0)
                {
                    /* LBL name */
                    if (scanner.lex() != TK_VALUE || *scanner.get_text() != '.')
                        goto syntax_error;
                    found_symbols[scanner.get_text()] = memory.size();
                }
                else if (strcmp(instn, "DB") == 0)
                {
                    /* DB val... ; */
                    int tokentype;
                    while ((tokentype = scanner.lex()) != TK_SEMICOLON)
                    {
                        if (tokentype != TK_VALUE)
                            goto syntax_error;
                        add_value(scanner.get_text());
                    }
                }
                else
                {
                    goto syntax_error;
                }
                break;
                
            default:
                goto syntax_error;
        }
    }
    
syntax_error:
    fprintf(stderr, "Syntax error on line %d.\n", scanner.get_lineno());
    return false;
}

bool assembler::link()
{
    bool result = true;
    
    for (need_sym_vec::const_iterator need_sym_iter = needed_symbols.begin(); need_sym_iter != needed_symbols.end(); ++need_sym_iter)
    {
        found_sym_map::const_iterator found = found_symbols.find(need_sym_iter->name);
        if (found != found_symbols.end())
        {
            memory[need_sym_iter->address] = found->second;
        }
        else
        {
            std::cerr << "Unresolved symbol '" << need_sym_iter->name << "'" << std::endl;
            result = false;
        }
    }
    
    return result;
}

void assembler::write_data(std::ostream& fout) const
{
    fout.write(reinterpret_cast<const char *>(&memory[0]), memory.size() * sizeof(uword_t));
}
