#include "ast.hh"
#include "dump.hh"
#include "machine.hh"

#include <cstdlib>
#include <iostream>
#include <map>
#include <set>
#include <string>
#include <vector>

typedef std::map<std::string, int> inttable;

struct compiler
{
    std::ostream          &f;
    const char            *curfunc;
    int                    nautos;
    int                    ntemps;
    int                    genid;
    inttable               locals;
    std::set<std::string>  extrns;
    std::set<std::string>  labels;
    std::vector<int>       switchstack;
    std::vector<int>       breakstack;
    bool                   error;
    
    compiler(std::ostream &f, const char *func) :
        f(f),
        curfunc(func),
        nautos(0),
        ntemps(0),
        genid(0),
        error(false)
    {
    }
    
    void prescan_expr(expression *expr);
    void prescan_stmt(statement *stmt);
    void put_lvalue_expression(expression *expr);
    void put_fun_args(expressions *args);
    void put_expression(expression *expr, bool logical_context = false);
    void put_switch_jumps(statement *stmt);
    void put_statement(statement *stmt);
};

static unsigned char escape_val(char c)
{
    switch (c)
    {
        case '0': return '\0';
        case 'e': return '\004'; /* EOT */
        case '(': return '{';
        case ')': return '}';
        case 't': return '\t';
        case '*': return '*';
        case '\'': return '\'';
        case '"': return '"';
        case 'n': return '\n';
        default: return 0;
    }
}

static uword_t charlit_val(const char *sval)
{
    uword_t result = 0;
    int shift = 0;
    unsigned char c;
    
    for (sval++; *sval != '\''; sval++, shift += CHAR_BITS)
    {
        if (*sval != '*')
            c = *sval;
        else
            c = escape_val(*++sval);
#ifdef B_ENDIAN_BIG
        result <<= CHAR_BITS;
        result += c;
#else
        result += (uword_t)c << shift;
#endif
    }
    return result;
}

static void put_string_val(std::ostream &f, const char *sval)
{
    uword_t wordval = 0;
    int i = 0;
    unsigned char c;
    
#ifdef B_ENDIAN_BIG
# define SHIFT ((WORD_CHARS - 1 - i) * CHAR_BITS)
#else
# define SHIFT (i * CHAR_BITS)
#endif
    f << "db ";
    while (*++sval != '"')
    {
        if (*sval != '*')
            c = *sval;
        else
            c = escape_val(*++sval);
        wordval += (uword_t)c << SHIFT;
        if (++i == WORD_CHARS)
        {
            f << wordval << "\n   ";
            wordval = 0;
            i = 0;
        }
    }
    wordval += (uword_t)'\004' << SHIFT;
    f << wordval << " ;\n";
#undef SHIFT
}

static void put_simple_definition(std::ostream &f, definition *def)
{
    f << "db ";
    if (!def->initialiser)
    {
        f << "0 ;\n";
    }
    else
    {
        expressions *exprs;
        int strings = 0;

        for (exprs = def->initialiser; exprs; exprs = exprs->next)
        {
            expression *expr = exprs->expr;
            switch (expr->type)
            {
                case EXP_INT:
                    f << expr->sval << "\n   ";
                    break;
                case EXP_CHAR:
                    f << charlit_val(expr->sval) << "\n   ";
                    break;
                case EXP_STRING:
                    f << "." << def->sval << "." << strings++ << "\n   ";
                    break;
                case EXP_NAME:
                    f << "." << expr->sval << "\n   ";
                    break;
                default:
                    std::cerr << "Internal error: bad initialiser " << expr->type << "." << std::endl;
                    std::exit(2);
            }
        }
        f << ";\n";
        if (strings)
        {
            strings = 0;
            for (exprs = def->initialiser; exprs; exprs = exprs->next)
            {
                expression *expr = exprs->expr;
                if (expr->type == EXP_STRING)
                {
                    f << "lbl ." << def->sval << "." << strings++ << "\n";
                    put_string_val(f, expr->sval);
                }
            }
        }
    }
}

static bool put_vector_definition(std::ostream &f, definition *def)
{
    bool result = true;
    
    uword_t len = 0;
    
    if (def->vlen)
    {
        switch (def->vlen->type)
        {
            case EXP_INT:
                len = atoword(def->vlen->sval);
                break;
            case EXP_CHAR:
                len = charlit_val(def->vlen->sval);
                break;
            case EXP_STRING:
                std::cerr << "Error: String literal as vector size." << std::endl;
                result = false;
                break;
            default:
                std::cerr << "Internal error: bad vector size expression " << def->vlen->type << "." << std::endl;
                exit(2);
        }
#if EXTENSIONS
        len++;
#endif
    }
    
    f << "db ." << def->sval << ".0 ;\n"
      << "lbl ." << def->sval << ".0\n"
      << "db ";
    
    expressions *exprs;
    int strings = 0;
    
    for (exprs = def->initialiser; exprs; exprs = exprs->next)
    {
        expression *expr = exprs->expr;
        switch (expr->type)
        {
            case EXP_INT:
                f << expr->sval << "\n   ";
                break;
            case EXP_CHAR:
                f << charlit_val(expr->sval) << "\n   ";
                break;
            case EXP_STRING:
                f << "." << def->sval << "." << ++strings << "\n   ";
                break;
            case EXP_NAME:
                f << "." << expr->sval << "\n   ";
                break;
            default:
                std::cerr << "Internal error: bad initialiser " << expr->type << "." << std::endl;
                exit(2);
        }
        if (len)
            len--;
    }
    while (len--)
        f << "0\n   ";
    f << ";\n";
    
    if (strings)
    {
        strings = 0;
        for (exprs = def->initialiser; exprs; exprs = exprs->next)
        {
            expression *expr = exprs->expr;
            if (expr->type == EXP_STRING)
            {
                f << "lbl ." << def->sval << "." << ++strings << "\n";
                put_string_val(f, expr->sval);
            }
        }
    }
    
    return result;
}

void compiler::prescan_expr(expression *expr)
{
    if (!expr)
        return;
    
    switch (expr->type)
    {
        case EXP_FUNCALL:
            for (expressions *args = expr->args; args; args = args->next)
                prescan_expr(args->expr);
            break;
            
        case EXP_STRING:
            expr->ival = ++genid;
            f << "lbl ." << curfunc << "." << expr->ival << "\n";
            put_string_val(f, expr->sval);
            break;
            
        default:
            break;
    }
    prescan_expr(expr->operands[0]);
    prescan_expr(expr->operands[1]);
    prescan_expr(expr->operands[2]);
}

void compiler::prescan_stmt(statement *stmt)
{
    for ( ; stmt; stmt = stmt->next)
    {
        switch (stmt->type)
        {
            case STMT_LABELLED:
                f << "lbl ." << curfunc << "." << stmt->sval << "\n";
                f << "db ." << curfunc << "." << stmt->sval << ".0 ;\n";
                labels.insert(stmt->sval);
                break;
                
            default:
                break;
        }
        prescan_expr(stmt->expr);
        prescan_stmt(stmt->contained[0]);
        prescan_stmt(stmt->contained[1]);
    }
}

void compiler::put_lvalue_expression(expression *expr)
{
    inttable::const_iterator nameval;
    
    switch (expr->type)
    {
        case EXP_UNARY:
            if (expr->ival == OP_DEREF)
            {
                put_expression(expr->operands[0]);
                f << "ld D A\n";
            }
            else
            {
                std::cerr << "Error: rvalue used as lvalue." << std::endl;
                error = true;
            }
            break;
            
        case EXP_NAME:
            if ((nameval = locals.find(expr->sval)) != locals.end())
            {
                if (nameval->second < 0)
                    f << "ld D S" << nameval->second << "\n";
                else
                    f << "ld D S+" << nameval->second << "\n";
            }
            else if (extrns.count(expr->sval))
            {
                f << "ld D ." << expr->sval << "\n";
            }
            else if (labels.count(expr->sval))
            {
                f << "ld D ." << curfunc << "." << expr->sval << "\n";
            }
            else
            {
#ifdef EXTENSIONS
                f << "ld D ." << expr->sval << "\n";
#else
                std::cerr << "Error: Undefined identifier '" << expr->sval << "'." << std::endl;
                error = true;
#endif
            }
            break;
            
        default:
            std::cerr << "Error: rvalue used as lvalue." << std::endl;
            error = true;
            break;
    }
}

void compiler::put_fun_args(expressions *args)
{
    if (!args)
        return;
    
    put_fun_args(args->next);
    
    put_expression(args->expr);
    f << "ld D S-" << (nautos + ++ntemps) << "\n";
    f << "ld (D) A\n";
}

void compiler::put_expression(expression *expr, bool logical_context /*= false*/)
{
    switch (expr->type)
    {
        case EXP_FUNCALL:
            put_fun_args(expr->args);
            put_expression(expr->operands[0]);
            f << "add S -" << (uword_t)(nautos + ntemps) << "\n";
            f << "call A\n";
            f << "add S " << (uword_t)(nautos + ntemps) << "\n";
            for (expressions *args = expr->args; args; args = args->next)
                ntemps--;
            break;
            
        case EXP_UNARY:
            switch (expr->ival)
            {
                case OP_POSTINC:
                    put_lvalue_expression(expr->operands[0]);
                    f << "ld A (D)\n";
                    f << "inc A\n";
                    f << "ld (D) A\n";
                    f << "dec A\n";
                    break;
                case OP_POSTDEC:
                    put_lvalue_expression(expr->operands[0]);
                    f << "ld A (D)\n";
                    f << "dec A\n";
                    f << "ld (D) A\n";
                    f << "inc A\n";
                    break;
                case OP_PREINC:
                    put_lvalue_expression(expr->operands[0]);
                    f << "ld A (D)\n";
                    f << "inc A\n";
                    f << "ld (D) A\n";
                    break;
                case OP_PREDEC:
                    put_lvalue_expression(expr->operands[0]);
                    f << "ld A (D)\n";
                    f << "dec A\n";
                    f << "ld (D) A\n";
                    break;
                case OP_NEGATE:
                    put_expression(expr->operands[0]);
                    f << "neg A\n";
                    break;
                case OP_NOT:
                    put_expression(expr->operands[0]);
                    f << "ld A Z\n";
                    break;
                case OP_COMP:
                    put_expression(expr->operands[0]);
                    f << "cpl A\n";
                    break;
                case OP_DEREF:
                    put_expression(expr->operands[0]);
                    f << "ld D A\n";
                    f << "ld A (D)\n";
                    break;
                case OP_ADDR:
                    put_lvalue_expression(expr->operands[0]);
                    f << "ld A D\n";
                    break;
                default:
                    std::cerr << "Internal error: bad unary operator " << expr->ival << "." << std::endl;
                    exit(2);
            }
            break;
            
        case EXP_BINARY:
#ifdef EXTENSIONS
            if (logical_context && (expr->ival == OP_AND || expr->ival == OP_OR))
            {
                int i = ++genid;
                put_expression(expr->operands[0], true);
                if (expr->ival == OP_AND)
                    f << "jz ." << curfunc << "." << i << "\n";
                else
                    f << "jnz ." << curfunc << "." << i << "\n";
                put_expression(expr->operands[1], true);
                f << "lbl ." << curfunc << "." << i << "\n";
            }
            else
#endif
            {
                put_expression(expr->operands[1]);
                f << "ld D S-" << (nautos + ++ntemps) << "\n";
                f << "ld (D) A\n";
                put_expression(expr->operands[0]);
                f << "ld D S-" << (nautos + ntemps--) << "\n";
                f << "ld D (D)\n";
                switch (expr->ival)
                {
                    case OP_MODULO: f << "mod A D\n"; break;
                    case OP_MULTIPLY: f << "mul A D\n"; break;
                    case OP_DIVIDE: f << "div A D\n"; break;
                    case OP_SUBTRACT: f << "sub A D\n"; break;
                    case OP_ADD: f << "add A D\n"; break;
                    case OP_LEFTSHIFT: f << "ls A D\n"; break;
                    case OP_RIGHTSHIFT: f << "rs A D\n"; break;
                    case OP_LESS: f << "lt A D\n"; break;
                    case OP_LESSEQUAL: f << "leq A D\n"; break;
                    case OP_GREATER: f << "gt A D\n"; break;
                    case OP_GREATEREQUAL: f << "geq A D\n"; break;
                    case OP_EQUAL: f << "eq A D\n"; break;
                    case OP_NOTEQUAL: f << "neq A D\n"; break;
                    case OP_AND: f << "and A D\n"; break;
                    case OP_XOR: f << "xor A D\n"; break;
                    case OP_OR: f << "or A D\n"; break;
                    default:
                        std::cerr << "Internal error: bad binary operator " << expr->ival << "." << std::endl;
                        exit(2);
                }
            }
            break;
            
        case EXP_ASSIGN:
            put_lvalue_expression(expr->operands[0]);
            f << "ld A D\n";
            f << "ld D S-" << (nautos + ++ntemps) << "\n";
            f << "ld (D) A\n";
            put_expression(expr->operands[1]);
            f << "ld D S-" << (nautos + ntemps--) << "\n";
            f << "ld D (D)\n";
            f << "ld (D) A\n";
            break;
            
        case EXP_COMPOUND:
            put_expression(expr->operands[1]);
            f << "ld D S-" << (nautos + ++ntemps) << "\n";
            f << "ld (D) A\n";
            put_lvalue_expression(expr->operands[0]);
            f << "ld A D\n";
            f << "ld D S-" << (nautos + ntemps+1) << "\n";
            f << "ld (D) A\n";
            f << "ld D A\n";
            f << "ld A (D)\n";
            f << "ld D S-" << (nautos + ntemps) << "\n";
            f << "ld D (D)\n";
            switch (expr->ival)
            {
                case OP_MODULO: f << "mod A D\n"; break;
                case OP_MULTIPLY: f << "mul A D\n"; break;
                case OP_DIVIDE: f << "div A D\n"; break;
                case OP_SUBTRACT: f << "sub A D\n"; break;
                case OP_ADD: f << "add A D\n"; break;
                case OP_LEFTSHIFT: f << "ls A D\n"; break;
                case OP_RIGHTSHIFT: f << "rs A D\n"; break;
                case OP_LESS: f << "lt A D\n"; break;
                case OP_LESSEQUAL: f << "leq A D\n"; break;
                case OP_GREATER: f << "gt A D\n"; break;
                case OP_GREATEREQUAL: f << "geq A D\n"; break;
                case OP_EQUAL: f << "eq A D\n"; break;
                case OP_NOTEQUAL: f << "neq A D\n"; break;
                case OP_AND: f << "and A D\n"; break;
                case OP_XOR: f << "xor A D\n"; break;
                case OP_OR: f << "or A D\n"; break;
                default:
                    std::cerr << "Internal error: bad compound operator " << expr->ival << "." << std::endl;
                    exit(2);
            }
            f << "ld D S-" << (nautos + ntemps+1) << "\n";
            f << "ld D (D)\n";
            f << "ld (D) A\n";
            ntemps--;
            break;
            
        case EXP_TERNARY:
        {
            int i = genid;
            genid += 2;
            put_expression(expr->operands[0], true);
            f << "jz ." << curfunc << "." << (i+1) << "\n";
            put_expression(expr->operands[1]);
            f << "jmp ." << curfunc << "." << (i+2) << "\n";
            f << "lbl ." << curfunc << "." << (i+1) << "\n";
            put_expression(expr->operands[2]);
            f << "lbl ." << curfunc << "." << (i+2) << "\n";
            break;
        }
            
        case EXP_INT:
            f << "ld A " << expr->sval << "\n";
            break;
            
        case EXP_CHAR:
            f << "ld A " << charlit_val(expr->sval) << "\n";
            break;
            
        case EXP_STRING:
            f << "ld A ." << curfunc << "." << expr->ival << "\n";
            break;
            
        case EXP_NAME:
            put_lvalue_expression(expr);
            f << "ld A (D)\n";
            break;
    }
}

void compiler::put_switch_jumps(statement *stmt)
{
    for ( ; stmt; stmt = stmt->next)
    {
        if (stmt->type == STMT_CASE)
        {
            uword_t i;
            switch (stmt->expr->type)
            {
                case EXP_INT:
                    i = atoword(stmt->expr->sval);
                    break;
                case EXP_CHAR:
                    i = charlit_val(stmt->expr->sval);
                    break;
                case EXP_STRING:
                    return;
                default:
                    std::cerr << "Internal error: bad case expression " << stmt->expr->type << "." << std::endl;
                    exit(2);
            }
            f << "ld A " << i << "\n";
            f << "sub A D\n";
            f << "jz ." << curfunc << "." << switchstack.back() << "." << i << "\n";
        }
        else if (stmt->type == STMT_SWITCH)
        {
            continue;
        }
        put_switch_jumps(stmt->contained[0]);
        put_switch_jumps(stmt->contained[1]);
    }
}

void compiler::put_statement(statement *stmt)
{
    int i;
    
    for ( ; stmt; stmt = stmt->next)
    {
        switch (stmt->type)
        {
            case STMT_AUTO:
                for (var_decl *decl = stmt->decls; decl; decl = decl->next)
                {
                    locals[decl->name] = - ++nautos;
                    if (decl->vlen)
                    {
                        f << "# auto " << decl->name << " " << decl->vlen->sval << "\n";
                        switch (decl->vlen->type)
                        {
                            case EXP_INT:
                                i = atoword(decl->vlen->sval);
                                break;
                            case EXP_CHAR:
                                i = charlit_val(decl->vlen->sval);
                                break;
                            case EXP_STRING:
                                std::cerr << "Error: auto vector size is string." << std::endl;
                                error = true;
                                return;
                            default:
                                std::cerr << "Internal error: bad auto vector size " << stmt->expr->type << "." << std::endl;
                                exit(2);
                        }
#if EXTENSIONS
                        i++;
#endif
                        f << "ld D S-" << (nautos + i) << "\n";
                        f << "ld A D\n";
                        f << "ld D S-" << nautos << "\n";
                        f << "ld (D) A\n";
                        nautos += i;
                    }
                }
                put_statement(stmt->contained[0]);
                break;
                
            case STMT_EXTRN:
                for (var_decl *decl = stmt->decls; decl; decl = decl->next)
                    extrns.insert(decl->name);
                put_statement(stmt->contained[0]);
                break;
                
            case STMT_LABELLED:
                f << "# " << stmt->sval << ":\n";
                f << "lbl ." << curfunc << "." << stmt->sval << ".0\n";
                put_statement(stmt->contained[0]);
                break;
                
            case STMT_CASE:
                if (switchstack.empty())
                {
                    std::cerr << "Error: case without switch." << std::cerr;
                    error = true;
                    break;
                }
                switch (stmt->expr->type)
                {
                    case EXP_INT:
                        f << "# case " << stmt->expr->sval << ":\n";
                        i = atoword(stmt->expr->sval);
                        break;
                    case EXP_CHAR:
                        f << "# case " << stmt->expr->sval << ":\n";
                        i = charlit_val(stmt->expr->sval);
                        break;
                    case EXP_STRING:
                        std::cerr << "Error: case constant is string." << std::endl;
                        error = true;
                    default:
                        std::cerr << "Internal error: bad case expression " << stmt->expr->type << "." << std::endl;
                        exit(2);
                }
                f << "lbl ." << curfunc << "." << switchstack.back() << "." << i << "\n";
                put_statement(stmt->contained[0]);
                break;
                
            case STMT_BLOCK:
                f << "# {\n";
                put_statement(stmt->contained[0]);
                f << "# }\n";
                break;
                
            case STMT_IF:
                f << "# if (";
                put_expression_source(f, stmt->expr, 0);
                f << ")\n";
                i = genid++;
                if (stmt->contained[1])
                    genid++;
                put_expression(stmt->expr, true);
                f << "jz ." << curfunc << "." << (i+1) << "\n";
                put_statement(stmt->contained[0]);
                if (stmt->contained[1])
                {
                    f << "jmp ." << curfunc << "." << (i+2) << "\n";
                    f << "# else\n";
                }
                f << "lbl ." << curfunc << "." << (i+1) << "\n";
                if (stmt->contained[1])
                {
                    put_statement(stmt->contained[1]);
                    f << "lbl ." << curfunc << "." << (i+2) << "\n";
                }
                break;
                
            case STMT_WHILE:
                f << "# while (";
                put_expression_source(f, stmt->expr, 0);
                f << ")\n";
                i = genid;
                genid += 2;
                f << "lbl ." << curfunc << "." << (i+1) << "\n";
                put_expression(stmt->expr, true);
                f << "jz ." << curfunc << "." << (i+2) << "\n";
                breakstack.push_back(i+2);
                put_statement(stmt->contained[0]);
                breakstack.pop_back();
                f << "jmp ." << curfunc << "." << (i+1) << "\n";
                f << "lbl ." << curfunc << "." << (i+2) << "\n";
                break;
                
            case STMT_BREAK:
                f << "# break;\n";
                if (breakstack.empty())
                {
                    std::cerr << "Error: break without while or switch." << std::endl;
                    error = true;
                }
                f << "jmp ." << curfunc << "." << breakstack.back() << "\n";
                break;
                
            case STMT_SWITCH:
                f << "# switch ";
                put_expression_source(f, stmt->expr, 0);
                f << "\n";
                put_expression(stmt->expr);
                f << "ld D A\n";
                switchstack.push_back(++genid);
                put_switch_jumps(stmt->contained[0]);
                f << "jmp ." << curfunc << "." << switchstack.back() << "\n";
                breakstack.push_back(switchstack.back());
                put_statement(stmt->contained[0]);
                f << "lbl ." << curfunc << "." << switchstack.back() << "\n";
                breakstack.pop_back();
                switchstack.pop_back();
                break;
                
            case STMT_GOTO:
                f << "# goto ";
                put_expression_source(f, stmt->expr, 0);
                f << ";\n";
                put_expression(stmt->expr);
                f << "jmp A\n";
                break;
                
            case STMT_RETURN:
                if (stmt->expr)
                {
                    f << "# return (";
                    put_expression_source(f, stmt->expr, 0);
                    f << ");\n";
                    put_expression(stmt->expr);
                }
                else
                {
                    f << "# return;\n";
                }
                f << "ret\n";
                break;
                
            case STMT_NULL:
                f << "# ;\n";
                break;
                
            case STMT_EXPRESSION:
                f << "# ";
                put_expression_source(f, stmt->expr, 0);
                f << ";\n";
                put_expression(stmt->expr);
                break;
        }
    }
}

static bool put_function_definition(std::ostream &f, definition *def)
{
    compiler compiler(f, def->sval);
    
    compiler.f << "db ." << compiler.curfunc << ".0 ;\n";
    compiler.prescan_stmt(def->stmt);
    
    int i = 1;
    for (var_decl *decl = def->args; decl; decl = decl->next)
        compiler.locals[decl->name] = i++;
    
    compiler.extrns.insert(def->sval);
    
    compiler.f << "lbl ." << compiler.curfunc << ".0\n";
    
    compiler.put_statement(def->stmt);
    
    compiler.f << "ret\n";
    
    return !compiler.error;
}

bool compile(definition *program, std::ostream &f)
{
    bool result = true;
    
    for (definition *def = program; def; def = def->next)
    {
        f << "lbl ." << def->sval << "\n";
        switch (def->type)
        {
            case DEF_SIMPLE:
                put_simple_definition(f, def);
                break;
            case DEF_VECTOR:
                if (!put_vector_definition(f, def))
                    result = false;
                break;
            case DEF_FUNCTION:
                if (!put_function_definition(f, def))
                    result = false;
                break;
        }
        f << "\n";
    }
    
    return result;
}
