#include "execute.hh"

#include <csetjmp>
#include <iomanip>
#include <iostream>

const uword_t instruction_size[MI__COUNT] =
{
#define INSTRUCTION(name, size) size,
#include "instructions.def"
};

#define OPSTACK_CAP 1024

struct execution_machine
{
    uword_t *memory;
    size_t   memory_size;
    
    /* Registers. */
    uword_t  a;
    uword_t  d;
    uword_t  sp;
    uword_t  pc;
    
    jmp_buf  jmpbuf;
    
    execution_machine(uword_t *mem, size_t size) :
        memory(mem), memory_size(size),
        a(0), d(0), sp(size - 1), pc(0),
        jmpbuf()
    {
    }
    
    uword_t mem_read(uword_t address);
    void mem_write(uword_t address, uword_t value);
    void trace(FILE *ftrace);
};

#define SIGNEXT(w) ((sword_t)((w) | ((w) & SWORD_MIN ? ~SWORD_MAX : 0)))

uword_t execution_machine::mem_read(uword_t address)
{
    if (address >= memory_size)
    {
        std::cerr << "Read invalid address 0" << std::oct << std::setfill('0') << std::setw(5) << address << "." << std::endl;
        std::longjmp(jmpbuf, 1);
    }
    
    return memory[address];
}

void execution_machine::mem_write(uword_t address, uword_t value)
{
    if (address >= memory_size)
    {
        std::cerr << "Write invalid address 0" << std::oct << std::setfill('0') << std::setw(5) << address << "." << std::endl;
        std::longjmp(jmpbuf, 1);
    }
    
    memory[address] = value;
}

void execution_machine::trace(FILE *ftrace)
{
    static const struct {
        const char *name;
        int size;
    } instructions[] = {
#define INSTRUCTION(name, size) { #name, size },
#include "instructions.def"
    };
    int i;
    
    fprintf(ftrace, "PC=%#.6"WORDFORMAT"o(", pc);
    if (pc >= memory_size)
    {
        fprintf(ftrace, "<>");
    }
    else
    {
        uword_t instn = memory[pc];
        if (instn >= MI__COUNT)
        {
            fprintf(ftrace, "%#.6"WORDFORMAT"o", instn);
        }
        else
        {
            fprintf(ftrace, "%s", instructions[instn].name);
            for (i = 1; i < instructions[instn].size; i++)
            {
                if (memory_size - pc <= (unsigned int)i)
                    fprintf(ftrace, " <>");
                else
                    fprintf(ftrace, " %#.6"WORDFORMAT"o", memory[pc + i]);
            }
        }
    }
    fprintf(ftrace, ") ");
    fprintf(ftrace, "A=%#.6"WORDFORMAT"o ", a);
    fprintf(ftrace, "D=%#.6"WORDFORMAT"o ", d);
    
    fprintf(ftrace, "Stack[%d:", (unsigned int)(memory_size - sp - 1));
    for (i = sp >= 4 ? -4 : (int)sp;
         i < (int)(memory_size - sp) && i <= 4;
         i++)
    {
        if (i == 0)
            fprintf(ftrace, " >%#.6"WORDFORMAT"o", memory[sp + i]);
        else
            fprintf(ftrace, " %#.6"WORDFORMAT"o", memory[sp + i]);
    }
    fprintf(ftrace, "]\n");
}

int execute(uword_t memory[], unsigned int memory_size, FILE *ftrace, int tracelevel)
{
    execution_machine machine(memory, memory_size);
    
    uword_t instn;
    uword_t next_addr;
    uword_t m, n;
    sword_t x, y, z;
    int i;
    
    if (setjmp(machine.jmpbuf))
    {
        machine.trace(stderr);
        return 0;
    }
    
    while (1)
    {
        if (tracelevel == TRACE_FULL)
            machine.trace(ftrace);
        
        instn = machine.mem_read(machine.pc);
        if (instn >= MI__COUNT)
        {
            fprintf(stderr, "Invalid instruction %#.6"WORDFORMAT"o at %#.6"WORDFORMAT"o.\n", instn, machine.pc);
            return 0;
        }
        machine.mem_read(machine.pc + instruction_size[instn] - 1); /* Check entire instruction in memory. */
        next_addr = machine.pc + instruction_size[instn];
        
        switch (instn)
        {
            case MI_NOP: /* NOP */
                break;
                
            case MI_HALT: /* HALT */
                return 1; /* Program halted normally. */
                
            case MI_JZ: /* JZ addr */
                if (machine.a == 0)
                    next_addr = machine.mem_read(machine.pc + 1);
                break;
                
            case MI_JNZ: /* JNZ addr */
                if (machine.a != 0)
                    next_addr = machine.mem_read(machine.pc + 1);
                break;
                
            case MI_JMP_I: /* JMP addr */
                next_addr = machine.mem_read(machine.pc + 1);
                break;
                
            case MI_JMP_A: /* JMP A */
                next_addr = machine.a;
                break;
                
            case MI_JMP_D: /* JMP D */
                next_addr = machine.d;
                break;
                
            case MI_CALL_I: /* CALL addr */
                if (tracelevel == TRACE_CALL)
                    machine.trace(ftrace);
                machine.mem_write(--machine.sp, next_addr);
                next_addr = machine.mem_read(machine.pc + 1);
                break;
                
            case MI_CALL_A: /* CALL A */
                if (tracelevel == TRACE_CALL)
                    machine.trace(ftrace);
                machine.mem_write(--machine.sp, next_addr);
                next_addr = machine.a;
                break;
                
            case MI_CALL_D: /* CALL D */
                if (tracelevel == TRACE_CALL)
                    machine.trace(ftrace);
                machine.mem_write(--machine.sp, next_addr);
                next_addr = machine.d;
                break;
                
            case MI_RET: /* RET */
                if (tracelevel == TRACE_CALL)
                    machine.trace(ftrace);
                next_addr = machine.mem_read(machine.sp++);
                break;
                
            case MI_LD_A_I: /* LD A, val */
                machine.a = machine.mem_read(machine.pc + 1);
                break;
                
            case MI_LD_A_D: /* LD A, D */
                machine.a = machine.d;
                break;
                
            case MI_LD_A_DD: /* LD A, (D) */
                machine.a = machine.mem_read(machine.d);
                break;
                
            case MI_LD_A_Z: /* LD A, Z */
                machine.a = machine.a == 0;
                break;
                
            case MI_LD_D_I: /* LD D, val */
                machine.d = machine.mem_read(machine.pc + 1);
                break;
                
            case MI_LD_D_A: /* LD D, A */
                machine.d = machine.a;
                break;
                
            case MI_LD_D_SI: /* LD D, S+off */
                machine.d = (machine.sp + machine.mem_read(machine.pc + 1)) & UWORD_MAX;
                break;
                
            case MI_LD_D_DD: /* LD D, (D) */
                machine.d = machine.mem_read(machine.d);
                break;
                
            case MI_LD_DD_A: /* LD (D), A */
                machine.mem_write(machine.d, machine.a);
                break;
                
            case MI_ADD: /* ADD A, D */
                n = (machine.a + machine.d) & UWORD_MAX;
                machine.a = n;
                break;
                
            case MI_ADD_S: /* ADD S, off */
                m = machine.mem_read(machine.pc + 1);
                n = (machine.sp + m) & UWORD_MAX;
                machine.sp = n;
                break;
                
            case MI_SUB: /* SUB A, D */
                n = (machine.a - machine.d) & UWORD_MAX;
                machine.a = n;
                break;
                
            case MI_MUL: /* MUL A, D */
                n = (machine.a * machine.d) & UWORD_MAX;
                machine.a = n;
                break;
                
            case MI_DIV: /* DIV A, D */
                x = SIGNEXT(machine.a);
                y = SIGNEXT(machine.d);
                z = x / y;
                machine.a = z & UWORD_MAX;
                break;
                
            case MI_MOD: /* MOD A, D */
                x = SIGNEXT(machine.a);
                y = SIGNEXT(machine.d);
                z = x % y;
                machine.a = z & UWORD_MAX;
                break;
                
            case MI_NEG: /* NEG A */
                machine.a = -machine.a & UWORD_MAX;
                break;
                
            case MI_INC: /* INC A */
                machine.a = (machine.a+1) & UWORD_MAX;
                break;
                
            case MI_DEC: /* DEC A */
                machine.a = (machine.a-1) & UWORD_MAX;
                break;
                
            case MI_LT: /* LT A D */
                x = SIGNEXT(machine.a);
                y = SIGNEXT(machine.d);
                machine.a = x < y;
                break;
                
            case MI_LEQ: /* LEQ A D */
                x = SIGNEXT(machine.a);
                y = SIGNEXT(machine.d);
                machine.a = x <= y;
                break;
                
            case MI_GT: /* GT A D */
                x = SIGNEXT(machine.a);
                y = SIGNEXT(machine.d);
                machine.a = x > y;
                break;
                
            case MI_GEQ: /* GEQ A D */
                x = SIGNEXT(machine.a);
                y = SIGNEXT(machine.d);
                machine.a = x >= y;
                break;
                
            case MI_EQ: /* EQ A D */
                x = SIGNEXT(machine.a);
                y = SIGNEXT(machine.d);
                machine.a = x == y;
                break;
                
            case MI_NEQ: /* NEQ A D */
                x = SIGNEXT(machine.a);
                y = SIGNEXT(machine.d);
                machine.a = x != y;
                break;
                
            case MI_AND: /* AND A, D */
                machine.a &= machine.d;
                break;
                
            case MI_OR: /* OR A, D */
                machine.a |= machine.d;
                break;
                
            case MI_XOR: /* XOR A, D */
                machine.a ^= machine.d;
                break;
                
            case MI_LS: /* LS A, D */
                machine.a = machine.a << machine.d & UWORD_MAX;
                break;
                
            case MI_RS: /* RS A, D */
                machine.a >>= machine.d;
                break;
                
            case MI_CPL: /* CPL A */
                machine.a = ~machine.a & UWORD_MAX;
                break;
                
            case MI_IN: /* IN A */
                i = getchar();
                machine.a = i == EOF ? '\004' : i;
                break;
                
            case MI_OUT: /* OUT A */
                n = machine.a;
                if (n == 0)
                {
                    putchar(0);
                }
                else
                {
                    for (i = 0; i < WORD_CHARS; i++)
                    {
#ifdef B_ENDIAN_BIG
                        x = (n >> (WORD_CHARS-1-i)*CHAR_BITS) & ((1<<CHAR_BITS) - 1);
#else
                        x = (n >> i*CHAR_BITS) & ((1<<CHAR_BITS) - 1);
#endif
                        if (x)
                            putchar(x);
                    }
                }
                break;
        }
        machine.pc = next_addr;
    }
}
